import os
from curses import beep
from datetime import datetime, timedelta, timezone
from math import ceil
from queue import Empty
from subprocess import DEVNULL, STDOUT, Popen
from sys import exit, stdout
from time import monotonic, time

from .events import (
    INPUT_END,
    INPUT_EXIT,
    INPUT_LAP,
    INPUT_MINUS,
    INPUT_PAUSE,
    INPUT_PLUS,
    INPUT_RESET,
)
from .ticker import Metronome
from .utils import (
    format_seconds,
    format_seconds_alt,
    format_target,
    parse_timestr,
)


def countdown(ui, args):
    target_time = parse_timestr(args.timespec)
    offset = (target_time.microsecond / 1_000_000)
    ticker = Metronome(ui.input_queue, offset=offset)
    ticker.start()

    while True:  # Outer loop to allow restarting countdown from scratch
        while True:  # Active countdown loop
            seconds_left = (target_time - datetime.now(timezone.utc)).total_seconds()
            if seconds_left <= 0:
                # If seconds_left is zero or negative, immediately break to handle the
                # "finished" state. This prevents displaying "0" for an entire second
                # while waiting for the next tick.
                break

            if args.alt_format:
                countdown_text = format_seconds_alt(
                    seconds_left, hide_seconds=args.no_seconds
                )
            else:
                countdown_text = format_seconds(
                    seconds_left, hide_seconds=args.no_seconds
                )

            with ui.curses_lock:
                ui.set_window_title(countdown_text)
                if args.outfile:
                    with open(args.outfile, "w") as f:
                        f.write("{}\n{}\n".format(countdown_text, seconds_left))
                end_text = (
                    format_target(
                        target_time,
                        time_format=args.time_format,
                        date_format=args.date_format,
                    )
                    if args.end
                    else None
                )
                color = 0
                if ticker.is_paused:
                    color = 3
                elif seconds_left <= args.critical:
                    color = 1
                ui.draw_text(countdown_text, color=color, end=end_text)

            annunciation = None
            if seconds_left <= args.critical:
                annunciation = str(int(ceil(seconds_left)))  # Announce whole seconds
            elif int(ceil(seconds_left)) in (5, 10, 20, 30, 60):
                annunciation = "{} {} seconds".format(
                    args.voice_prefix, int(ceil(seconds_left))
                )
            elif int(ceil(seconds_left)) in (300, 600, 1800):
                annunciation = "{} {} minutes".format(
                    args.voice_prefix, int(seconds_left / 60)
                )
            elif int(ceil(seconds_left)) == 3600:
                annunciation = "{} one hour".format(args.voice_prefix)
            if annunciation or args.exec_cmd:
                if (
                    annunciation and args.voice_cmd
                ):  # Only announce if there is something to say
                    Popen(
                        [args.voice_cmd, "-v", args.voice, annunciation.strip()],
                        stdout=DEVNULL,
                        stderr=STDOUT,
                    )
                if args.exec_cmd:
                    # Pass annunciation even if it's empty, format() handles it.
                    Popen(
                        args.exec_cmd.format(seconds_left, annunciation or ""),
                        stdout=DEVNULL,
                        stderr=STDOUT,
                        shell=True,
                    )

            input_action = ui.input_queue.get()
            if input_action == INPUT_PAUSE:
                duration = ticker.pause()
                if duration:
                    target_time += timedelta(seconds=duration)
            elif input_action == INPUT_EXIT:
                exit(1)
            elif input_action == INPUT_RESET:
                target_time = parse_timestr(args.timespec)
                # Continue the inner loop, seconds_left will be re-evaluated
                continue
            elif input_action == INPUT_PLUS:
                target_time += timedelta(seconds=10)
            elif input_action == INPUT_MINUS:
                target_time -= timedelta(seconds=10)
            elif input_action == INPUT_END:
                args.end = not args.end
                continue

        # After the active countdown loop, handle the "time is up" state.

        if not args.no_bell:
            with ui.curses_lock:
                beep()

        if args.outfile:
            with open(args.outfile, "w") as f:
                f.write("{}\n{}\n".format(args.text if args.text else "DONE", 0))

        if args.blink or args.text:
            base_color = 1 if args.blink else 0
            flip = True
            ticker.pause()  # Pause ticker during blinking phase
            while True:
                if args.quit_after and (
                    datetime.now(timezone.utc) - target_time
                ).total_seconds() > float(args.quit_after):
                    return
                with ui.curses_lock:
                    ui.set_window_title("/" if flip else "\\")
                    if args.text:
                        ui.draw_text(args.text, color=base_color if flip else 4)
                    else:
                        ui.draw_text("", color=base_color if flip else 4)
                if args.blink:
                    flip = not flip
                try:
                    input_action = ui.input_queue.get(True, 0.5)
                except Empty:
                    input_action = None
                if input_action == INPUT_EXIT:
                    return
                elif input_action == INPUT_RESET:
                    target_time = parse_timestr(args.timespec)
                    ticker.pause()  # resume
                    break  # Break out of the blinking loop to restart the main countdown
        else:
            break


def clock(ui, args):
    time_started = monotonic()
    seconds_elapsed = 0
    offset = timedelta(0)
    ticker = Metronome(ui.input_queue)
    ticker.start()
    while True:
        seconds_elapsed = monotonic() - time_started
        if args.quit_after and seconds_elapsed >= float(args.quit_after):
            return
        clock_text = (datetime.now() + offset).strftime(args.time_format)
        with ui.curses_lock:
            ui.set_window_title(clock_text)

            if args.outfile:
                with open(args.outfile, "w") as f:
                    f.write("{}\n{}\n".format(clock_text, seconds_elapsed))

            ui.draw_text(clock_text, color=3 if ticker.is_paused else 0)

        input_action = ui.input_queue.get()
        if input_action == INPUT_EXIT:
            break
        if input_action == INPUT_PLUS:
            offset += timedelta(seconds=10)
        elif input_action == INPUT_MINUS:
            offset -= timedelta(seconds=10)
        elif input_action == INPUT_PAUSE:
            ticker.pause()
        elif input_action == INPUT_RESET:
            offset = timedelta(0)


def stopwatch(ui, args):
    time_started = monotonic()
    ticker = Metronome(ui.input_queue, offset=time() % 1.0)
    ticker.start()
    time_paused = None
    seconds_elapsed = 0
    laps = []
    while True:
        if not time_paused:
            seconds_elapsed = monotonic() - time_started
        else:
            seconds_elapsed = time_paused - time_started

        if args.quit_after and seconds_elapsed >= float(args.quit_after):
            return seconds_elapsed, laps

        if args.alt_format:
            stopwatch_text = format_seconds_alt(
                round(seconds_elapsed), hide_seconds=args.no_seconds
            )
        else:
            stopwatch_text = format_seconds(
                round(seconds_elapsed), hide_seconds=args.no_seconds
            )
        with ui.curses_lock:
            if not args.no_window_title:
                os.write(
                    stdout.fileno(),
                    "\033]2;{0}\007".format(stopwatch_text).encode(),
                )
            if args.outfile:
                with open(args.outfile, "w") as f:
                    f.write("{}\n{}\n".format(stopwatch_text, seconds_elapsed))

            ui.draw_text(stopwatch_text, color=3 if ticker.is_paused else 0)

        annunciation = None
        if int(seconds_elapsed) <= args.critical and seconds_elapsed >= 1:
            annunciation = str(int(seconds_elapsed))
        elif int(seconds_elapsed) in (5, 10, 20, 30, 40, 50, 60):
            annunciation = "{} {} seconds".format(
                args.voice_prefix, int(seconds_elapsed)
            )
        elif int(seconds_elapsed) in (120, 180, 300, 600, 1800):
            annunciation = "{} {} minutes".format(
                args.voice_prefix, int(seconds_elapsed / 60)
            )
        elif int(seconds_elapsed) == 3600:
            annunciation = "{} one hour".format(args.voice_prefix)
        elif int(seconds_elapsed) % 3600 == 0 and seconds_elapsed > 1:
            annunciation = "{} {} hours".format(
                args.voice_prefix, int(seconds_elapsed / 3600)
            )

        if annunciation and args.voice_cmd:
            Popen(
                [args.voice_cmd, "-v", args.voice, annunciation.strip()],
                stdout=DEVNULL,
                stderr=STDOUT,
            )
        if args.exec_cmd:
            Popen(
                args.exec_cmd.format(seconds_elapsed, annunciation or ""),
                stdout=DEVNULL,
                stderr=STDOUT,
                shell=True,
            )

        input_action = ui.input_queue.get()
        if input_action == INPUT_PLUS:
            time_started -= 10
        elif input_action == INPUT_MINUS:
            time_started += 10
        elif input_action == INPUT_PAUSE:
            duration = ticker.pause()
            if duration:  # unpaused
                time_started += duration
                time_paused = None
            else:
                time_paused = monotonic()
        elif input_action == INPUT_EXIT:
            break
        elif input_action == INPUT_RESET:
            laps = []
            time_started = monotonic()
        elif input_action == INPUT_LAP:
            lap_time = monotonic()
            laps.append(lap_time - time_started)
            time_started = lap_time

    return (monotonic() - time_started, laps)



from math import ceil
from threading import Thread
from time import monotonic, sleep, time

from .events import TIME_TICK


class Metronome:
    """
    Will put a TIME_TICK event into the given queue every full second, delayed by offset.
    """

    def __init__(self, queue, offset=0):
        self._offset = offset
        self._queue = queue
        self._thread = Thread(target=self._run, daemon=True)
        self._pause_time = None

    def _run(self):
        while True:
            # Calculate the time to sleep until the next full second
            current_time = time()
            target_time = ceil(current_time) + self._offset
            sleep_time = target_time - current_time

            # If sleep_time is very small (e.g., we're just before the second),
            # we should wait for the next second.
            if sleep_time < 0.001:  # Small buffer to avoid missing the second
                sleep_time += 1.0

            sleep(sleep_time)
            if not self.is_paused:
                self._queue.put(TIME_TICK)

    def start(self):
        self._thread.start()

    def pause(self):
        if self.is_paused:  # unpause
            duration = monotonic() - self._pause_time
            self._pause_time = None
            if self._offset is not None:
                # If we're not running in clock mode, change the offset so the
                # next tick will be 1s from now.
                self._offset = time() % 1.0
            return duration
        else:  # pause
            self._pause_time = monotonic()

    @property
    def is_paused(self):
        return self._pause_time is not None



import curses
import os
from queue import Queue
from sys import stdout
from threading import Lock, Thread
from time import sleep

from art import text2art

from .events import (
    INPUT_END,
    INPUT_EXIT,
    INPUT_LAP,
    INPUT_MINUS,
    INPUT_PAUSE,
    INPUT_PLUS,
    INPUT_RESET,
)
from .ttf import ttf_to_ascii
from .utils import pad_to_size


class Ui:
    def __init__(self, stdscr, args):
        self.curses_lock = Lock()
        self.input_queue = Queue()
        self.stdscr = stdscr
        self._args = args

        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_RED, -1)
        curses.init_pair(2, curses.COLOR_RED, curses.COLOR_RED)
        curses.init_pair(3, curses.COLOR_BLUE, -1)
        curses.init_pair(4, -1, curses.COLOR_RED)
        try:
            curses.curs_set(False)
        except curses.error:
            # fails on some terminals
            pass
        stdscr.timeout(0)  # Set timeout for getch/getkey to 0 (non-blocking)
        stdscr.nodelay(True)  # Set nodelay mode (also non-blocking)

    def draw_text(self, text, color=0, end=None):
        """
        Draws text in the given color. Duh.
        """

        title = self._args.title or ""
        end = end or ""

        # build a list of possible fallbacks in descending order of preference
        variants = [
            (title + "\n\n" + text + "\n\n" + end).strip("\n"),
            (title + "\n\n" + text).strip("\n"),
            text,
            "E",
        ]

        if not self._args.no_art:
            if os.path.exists(self._args.font):
                ttf_args = (
                    self._args.font,
                    self._args.font_size,
                    self._args.font_charset,
                )
                art_title = ttf_to_ascii(title, *ttf_args)
                art_text = ttf_to_ascii(text, *ttf_args)
                art_end = ttf_to_ascii(end, *ttf_args)
            else:
                art_title = text2art(title, font=self._args.font)
                art_text = text2art(text, font=self._args.font)
                art_end = text2art(end, font=self._args.font)
            variants = [
                (art_title + "\n\n\n\n" + art_text + "\n\n\n\n" + art_end).strip("\n"),
                (art_title + "\n\n" + art_text + "\n\n" + art_end).strip("\n"),
                (art_title + "\n\n" + art_text + "\n\n" + end).strip("\n"),
                (title + "\n\n" + art_text + "\n\n" + end).strip("\n"),
            ] + variants

        y, x = self.stdscr.getmaxyx()
        for variant in variants:
            lines = pad_to_size(variant, x, y).rstrip("\n").split("\n")
            self.stdscr.erase()
            try:
                for i, line in enumerate(lines):
                    self.stdscr.insstr(i, 0, line, curses.color_pair(color))
            except Exception:
                continue
            else:
                break

        self.stdscr.refresh()

    def set_window_title(self, text):
        if not self._args.no_window_title:
            os.write(
                stdout.fileno(),
                "\033]2;{0}\007".format(text).encode(),
            )

    def start_input_thread(self):
        Thread(
            target=self._input_thread_body,
            daemon=True,
        ).start()

    def _input_thread_body(self):
        while True:
            try:
                with self.curses_lock:
                    key = self.stdscr.getkey()
            except Exception:
                key = None
            if key in ("q", "Q"):
                self.input_queue.put(INPUT_EXIT)
            elif key == " ":
                self.input_queue.put(INPUT_PAUSE)
            elif key in ("e", "E"):
                self.input_queue.put(INPUT_END)
            elif key in ("r", "R"):
                self.input_queue.put(INPUT_RESET)
            elif key in ("l", "L"):
                self.input_queue.put(INPUT_LAP)
            elif key == "+":
                self.input_queue.put(INPUT_PLUS)
            elif key == "-":
                self.input_queue.put(INPUT_MINUS)
            sleep(0.01)



import os
from argparse import ArgumentParser, RawTextHelpFormatter
from curses import wrapper
from functools import wraps
from os.path import abspath, dirname
from sys import stderr

from . import VERSION
from .modes import clock, countdown, stopwatch
from .ui import Ui
from .utils import format_seconds, normalize_text

DEFAULT_FONT = "univers"
DEFAULT_TIME_FORMAT = "%H:%M:%S"  # --no-seconds expects this to end with :%S
DEFAULT_DATE_FORMAT = "%Y-%m-%d"


def _escape_percent_for_argparse_help(text):
    """Escapes literal percent signs in text for argparse help messages."""
    return text.replace("%", "%%")


def graceful_ctrlc(func):
    """
    Makes the decorated function exit with code 1 on CTRL+C.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            exit(1)

    return wrapper


parser = ArgumentParser(
    prog="termdown",
    description="""
    Starts a countdown to TIME. Example values for TIME:
    10, '1h 5m 30s', '12:00', '2020-01-01', '2020-01-01 14:00 UTC'.
    \b
    If TIME is not given, termdown will operate in stopwatch mode
    and count forward.
    \b
    Hotkeys:
    \tE\tShow end time (countdown mode only)
    \tL\tLap (stopwatch mode only)
    \tR\tReset
    \tSPACE\tPause (will delay absolute TIME)
    \t+\tPlus (will add 10 seconds)
    \t-\tMinus (will subtract 10 seconds)
    \tQ\tQuit""",
    epilog="Tempus fugit.",
    formatter_class=RawTextHelpFormatter,
)
parser.add_argument(
    "-a",
    "--alt-format",
    action="store_true",
    help="Use colon-separated time format",
)
parser.add_argument(
    "-b", "--blink", action="store_true", help="Flash terminal at end of countdown"
)
parser.add_argument(
    "-B",
    "--no-bell",
    action="store_true",
    help="Don't ring terminal bell at end of countdown",
)
parser.add_argument(
    "-c",
    "--critical",
    type=int,
    default=3,
    metavar="N",
    help="Draw final N seconds in red and announce them individually with --voice "
    "or --exec-cmd (defaults to 3)",
)
parser.add_argument(
    "-e",
    "--end",
    action="store_true",
    help="Display target datetime of unpaused countdown",
)
parser.add_argument(
    "-f",
    "--font",
    default=DEFAULT_FONT,
    metavar="FONT",
    help="Choose from https://www.ascii-art.site/FontList.html or provide a full path to an OTF/TTF file",
)
parser.add_argument(
    "--font-charset",
    default=" ░▒▓█",
    metavar="CHARSET",
    help='Provide a string of characters of increasing visual density (e.g. " .oO#@") to render OTF/TTF pixels',
)
parser.add_argument(
    "--font-size",
    type=int,
    default=24,
    metavar="N",
    help="Set font size when using OTF/TTF",
)
parser.add_argument(
    "-p",
    "--voice-prefix",
    metavar="TEXT",
    help="Add TEXT to the beginning of --voice and --exec annunciations "
    "(except per-second ones)",
)
parser.add_argument(
    "-q",
    "--quit-after",
    type=int,
    metavar="N",
    help="Quit N seconds after countdown (use with -b or -t) "
    "or terminate stopwatch after N seconds",
)
parser.add_argument(
    "-s",
    "--no-seconds",
    action="store_true",
    help="Don't show seconds (except for last minute of countdown "
    "and first minute of stopwatch)",
)
parser.add_argument("-t", "--text", help="Text to display at end of countdown")
parser.add_argument(
    "-T", "--title", help="Text to display on top of countdown/stopwatch"
)
parser.add_argument(
    "-W",
    "--no-window-title",
    action="store_true",
    help="Don't update terminal title with remaining/elapsed time",
)
parser.add_argument(
    "-v",
    "--voice",
    metavar="VOICE",
    help="Spoken countdown "
    "(at fixed intervals with per-second annunciations starting at --critical; "
    "requires `espeak` on Linux or `say` on macOS; "
    "choose VOICE from `say -v '?'` or `espeak --voices`)",
)
parser.add_argument(
    "-o",
    "--outfile",
    metavar="PATH",
    help="File to write current remaining/elapsed time to",
)
parser.add_argument(
    "--exec-cmd",
    metavar="CMD",
    help="Runs CMD every second. '{0}' and '{1}' in CMD will be replaced with the "
    "remaining/elapsed number of seconds and a more sparse annunciation as in "
    "--voice, respectively. For example, to get a callout at five seconds only, "
    "use: --exec-cmd \"if [ '{0}' == '5' ]; then say -v Alex {1}; fi\"",
)
parser.add_argument(
    "--no-art", action="store_true", help="Don't use ASCII art for display"
)
parser.add_argument(
    "--no-text-magic",
    action="store_true",
    help="Don't try to replace non-ASCII characters (use with -t)",
)
parser.add_argument(
    "-z",
    "--time",
    action="store_true",
    help="Show current time instead of countdown/stopwatch",
)
parser.add_argument(
    "-Z",
    "--time-format",
    default=None,
    help=f'Format for --time/--end (defaults to "{_escape_percent_for_argparse_help(DEFAULT_TIME_FORMAT)}", ignores --no-seconds)',
)
parser.add_argument(
    "-D",
    "--date-format",
    default=None,
    help=f'Format for --end (defaults to "{_escape_percent_for_argparse_help(DEFAULT_DATE_FORMAT)}")',
)
parser.add_argument(
    "timespec",
    nargs="?",
    default=None,
    help="TIME to countdown to. Example values: 10, '1h 5m 30s', '12:00', "
    "'2020-01-01', '2020-01-01 14:00 UTC'. If not given, operates in stopwatch mode.",
)
parser.add_argument(
    "--version",
    action="version",
    version=f"%(prog)s {VERSION}",
    help="Show version and exit",
)


@graceful_ctrlc
def curses_ui(stdscr, mode, args):
    ui = Ui(stdscr, args)
    ui.start_input_thread()
    try:
        return mode(ui, args)
    finally:
        with ui.curses_lock:
            if not args.no_window_title:
                ui.set_window_title("")


def main():
    args = parser.parse_args()
    if args.exec_cmd and args.voice:  # prevent passing both --exec-cmd and --voice
        raise RuntimeError("--exec-cmd and --voice are mutually exclusive")

    if args.time_format is None:
        args.time_format = (
            DEFAULT_TIME_FORMAT[:-3] if args.no_seconds else DEFAULT_TIME_FORMAT
        )
    if args.date_format is None:
        args.date_format = DEFAULT_DATE_FORMAT

    if args.outfile:
        if os.path.exists(args.outfile):
            raise RuntimeError("File already exists: {}".format(args.outfile))
        if not os.access(dirname(abspath(args.outfile)), os.W_OK):
            raise RuntimeError("Unable to write file: {}".format(args.outfile))

    if args.text and not args.no_text_magic:
        args.text = normalize_text(args.text)

    args.voice_cmd = None
    if args.voice:
        for cmd in ("/usr/bin/say", "/usr/bin/espeak", "/usr/bin/espeak-ng"):
            if os.path.exists(cmd):
                args.voice_cmd = cmd
                break
    if args.voice or args.exec_cmd:
        args.voice_prefix = args.voice_prefix or ""

    if args.time:
        wrapper(curses_ui, clock, args)
    elif args.timespec:
        wrapper(curses_ui, countdown, args)
    else:
        seconds_elapsed, laps = wrapper(curses_ui, stopwatch, args)

        for lap_index, lap_time in enumerate(laps):
            stderr.write(
                "{:.3f}\t{}\tlap {}\n".format(
                    lap_time,
                    format_seconds(int(lap_time)),
                    lap_index + 1,
                )
            )

        if laps:
            stderr.write(
                "{:.3f}\t{}\tlap {}\n".format(
                    seconds_elapsed,
                    format_seconds(int(seconds_elapsed)),
                    len(laps) + 1,
                )
            )
            laps.append(seconds_elapsed)
            total_seconds = sum(laps)
            average_seconds = total_seconds / len(laps)
            stderr.write(
                "{:.3f}\t{}\tlap min\n".format(
                    min(laps),
                    format_seconds(int(min(laps))),
                )
            )
            stderr.write(
                "{:.3f}\t{}\tlap max\n".format(
                    max(laps),
                    format_seconds(int(max(laps))),
                )
            )
            stderr.write(
                "{:.3f}\t{}\tlap avg\n".format(
                    average_seconds,
                    format_seconds(int(average_seconds)),
                )
            )
            stderr.write(
                "{:.3f}\t{}\ttotal\n".format(
                    total_seconds,
                    format_seconds(int(total_seconds)),
                )
            )
        else:
            stderr.write(
                "{:.3f}\t{}\ttotal\n".format(
                    seconds_elapsed,
                    format_seconds(int(seconds_elapsed)),
                )
            )
        stderr.flush()



import re
import unicodedata
from datetime import datetime, timedelta, timezone
from math import ceil

from dateutil import tz
from dateutil.parser import parse

NORMALIZE_TEXT_MAP = {
    "ä": "ae",
    "Ä": "Ae",
    "ö": "oe",
    "Ö": "Oe",
    "ü": "ue",
    "Ü": "Ue",
    "ß": "ss",
}
TIMEDELTA_REGEX = re.compile(
    r"((?P<years>\d+)y ?)?"
    r"((?P<days>\d+)d ?)?"
    r"((?P<hours>\d+)h ?)?"
    r"((?P<minutes>\d+)m ?)?"
    r"((?P<seconds>\d+)s ?)?"
)
DATE_COMPONENT_REGEX = re.compile(
    r"\d{4}|"  # year
    r"jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec|"
    r"\d{1,2}[/\-]\d{1,2}",  # day/month with separator
    re.IGNORECASE,
)


def format_seconds(seconds, hide_seconds=False):
    """
    Returns a human-readable string representation of the given amount
    of seconds.
    """
    seconds = int(ceil(seconds))
    if seconds <= 60:
        return str(seconds)
    output = ""
    for period, period_seconds in (
        ("y", 31557600),
        ("d", 86400),
        ("h", 3600),
        ("m", 60),
        ("s", 1),
    ):
        if seconds >= period_seconds and not (hide_seconds and period == "s"):
            output += str(int(seconds / period_seconds))
            output += period
            output += " "
            seconds = seconds % period_seconds
    return output.strip()


def format_seconds_alt(seconds, hide_seconds=False):
    output = ""
    seconds = int(ceil(seconds))
    total_seconds = seconds
    for period_seconds in (
        31557600,
        86400,
        3600,
        60,
        1,
    ):
        if hide_seconds and period_seconds == 1 and total_seconds > 60:
            break
        actual_period_value = int(seconds / period_seconds)
        if actual_period_value > 0:
            output += str(actual_period_value).zfill(2) + ":"
        elif 86400 > period_seconds or total_seconds > period_seconds:
            output += "00:"
        seconds = seconds % period_seconds
    return output.rstrip(":")


def format_target(target, time_format, date_format):
    """
    Returns a human-readable string representation of the countdown's target
    datetime. Converts UTC target to local time for display.
    """
    target = target.astimezone(tz.tzlocal())
    if datetime.now().date() != target.date():
        fmt = "{} {}".format(date_format, time_format)
    else:
        fmt = time_format
    return target.strftime(fmt)


def normalize_text(input_str):
    for char, replacement in NORMALIZE_TEXT_MAP.items():
        input_str = input_str.replace(char, replacement)
    return "".join(
        [
            c
            for c in unicodedata.normalize("NFD", input_str)
            if unicodedata.category(c) != "Mn"
        ]
    )


def pad_to_size(text, x, y):
    """
    Adds whitespace to text to center it within a frame of the given
    dimensions.
    """
    input_lines = text.rstrip("\n").split("\n")
    number_of_input_lines = len(input_lines)
    output = ""

    padding_top = int((y - number_of_input_lines) / 2)
    padding_bottom = y - number_of_input_lines - padding_top

    output += padding_top * (" " * x + "\n")
    for line in input_lines:
        padding_left = int((x - len(line)) / 2)
        output += (
            padding_left * " " + line + " " * (x - padding_left - len(line)) + "\n"
        )
    output += padding_bottom * (" " * x + "\n")
    return output


def parse_timestr(timestr):
    """
    Parse a string describing a point in time.
    Returns a timezone-aware datetime in UTC to avoid issues with DST changes.
    """
    timedelta_secs = parse_timedelta(timestr)

    if timedelta_secs:
        target = datetime.now(timezone.utc) + timedelta(seconds=timedelta_secs)
    elif timestr.isdigit():
        target = datetime.now(timezone.utc) + timedelta(seconds=int(timestr))
    else:
        try:
            target = parse(timestr)
        except Exception:
            # unfortunately, dateutil doesn't raise the best exceptions
            raise ValueError("Unable to parse '{}'".format(timestr))
        if target.tzinfo is None:
            target = target.replace(tzinfo=tz.tzlocal())
        target = target.astimezone(timezone.utc)

        if (
            target <= datetime.now(timezone.utc) and
            not DATE_COMPONENT_REGEX.search(timestr)
        ):
            # User only gave us a time and it's in the past - probably
            # means that time tomorrow.
            target += timedelta(days=1)

    return target


def parse_timedelta(deltastr):
    """
    Parse a string describing a period of time.
    """
    matches = TIMEDELTA_REGEX.match(deltastr)
    if not matches:
        return None
    components = {}
    for name, value in matches.groupdict().items():
        if value:
            components[name] = int(value)
    for period, hours in (("days", 24), ("years", 8766)):
        if period in components:
            components["hours"] = (
                components.get("hours", 0) + components[period] * hours
            )
            del components[period]
    return int(timedelta(**components).total_seconds())



from PIL import Image, ImageDraw, ImageFont


def ttf_to_ascii(text, font_path, font_size, char_set):
    try:
        font = ImageFont.truetype(font_path, font_size)
    except IOError:
        raise RuntimeError("Error: Could not load font from {font_path}")

    # Figure out text dimensions
    dummy_img = Image.new("L", (1, 1))
    dummy_draw = ImageDraw.Draw(dummy_img)
    bbox = dummy_draw.textbbox((0, 0), text, font=font)
    width = bbox[2] - bbox[0]
    height = bbox[3] - bbox[1]

    if width == 0 or height == 0:
        return ""

    # Create actual image
    img = Image.new("L", (width, height), color=255)
    draw = ImageDraw.Draw(img)
    draw.text((-bbox[0], -bbox[1]), text, font=font, fill=0)

    pixels = img.getdata()
    lines = []
    char_range = len(char_set) - 1

    for i in range(height):
        line = []
        for j in range(width):
            pixel_value = pixels[i * width + j]
            # Invert the pixel value because 0 is black and 255 is white.
            # We want darker pixels to map to denser ASCII characters.
            inverted_pixel = 255 - pixel_value
            index = int(inverted_pixel / 255 * char_range)
            line.append(char_set[index])
        lines.append("".join(line))

    return "\n".join(lines)
